/*!
 * Copyright (c) 2020 by Contributors
 * \file base.h
 * \brief Macros common to all headers
 *
 * \author Hyunsu Cho
 */

#ifndef RABIT_BASE_H_
#define RABIT_BASE_H_

#ifndef _CRT_SECURE_NO_WARNINGS
#define _CRT_SECURE_NO_WARNINGS
#endif  // _CRT_SECURE_NO_WARNINGS
#ifndef _CRT_SECURE_NO_DEPRECATE
#define _CRT_SECURE_NO_DEPRECATE
#endif  // _CRT_SECURE_NO_DEPRECATE

#endif  // RABIT_BASE_H_
